{% extends "djhcup_staging_base.html" %}
{% block title %}Inventory Summary{% endblock %}
{% block body %}
<p><a href="/">&larr;Hachoir Index</a></p>
<p><a href="/staging">&larr;Staging Index</a></p>
<h1>{% block heading %}{% if title %}{{ title }}{% else %}Inventory Summary{% endif %}{% endblock %}</h1>


{% if counts %}
    <ul>
    {% for c in counts %}
        <li>{{ c.name }}: {{ c.value }}</li>
    {% endfor %}
    </ul>
{% endif %}


{% if links %}
    <ul>
    {% for l in links %}
        <li><a href="{{ l.href }}">{{ l.display }}</a></li>
    {% endfor %}
    </ul>
{% endif %}


{% if objects %}
    <h2>{% block object_title %}Objects{% endblock %}</h2>
    <ul>
    {% for o in objects %}
        <li><a href="{{ o.get_absolute_url }}">{{ o }} ({{ o.pk }})</a></li>
    {% endfor %}
    </ul>
{% endif %}


{% endblock %}
