{% extends "djhcup_staging_base.html" %}

{% block title %}Hachoir - Details for {{ object }}{% endblock %}

{% block body %}
<p><a href="/">&larr;Hachoir Index</a></p>
<p><a href="/staging">&larr;Staging Index</a></p>

<h1>{{ object }}</h1>


{% if counts %}
<ul>
{% for c in counts %}
<li>{{ c.name }}: {{ c.value }}</li>
{% endfor %}
</ul>
{% endif %}

{% if object.parent %}
<p>Parent: <a href="{{ object.parent.get_absolute_url}}">{{ object.parent }}</a></p>
{% endif %}


{% if children %}
<h2>Children of this object</h2>
<ul>
{% for c in children %}
<li><a href="{{ c.get_absolute_url }}">{{ c }}</a></li>
{% endfor %}
</ul>
{% endif %}


{% block details %}


{% endblock %}
{% endblock %}
