{% extends "djhcup_base.html" %}

{% block title %}
{% if title %}{{ title }}{% else %}No Title{% endif %}
{% endblock %}

{% block head %}
{% endblock %}

{% block body %}
{% if title %}<h1>{{ title }}</h1>{% endif %} 
{% if response_messages %}
<h2>Response Messages</h2>
<ol class="response_messages">
{% for m in response_messages %}
    <li class="{{ m.type }}">{{ m.content }}</li>
{% endfor %}
</ol>
{% elif links %}
<ul class="links">
{% for l in links %}
    <li><a href="{{ l.href }}">{{ l.display }}</a></li>
{% endfor %}
</ul>
{% else %}<p>No content.</p>{% endif %}
{% endblock %}
</body>
</html>